module MemoryController (CLK, RESET, DATAIN, ENABLE, NO, NEWPAC, MEMWRITE, DOUT, ADDRO);

input CLK ;		// RXCLK	
input RESET;		// External Reset
input [3:0] DATAIN;	// INPUT from PHYS
input ENABLE;		// WRITE THIS DATA
input NO;		// This packet is crap, don't keep it

output NEWPAC;		// Signal external control that we have a new, good, packet
output MEMWRITE;	// Write to memory (hold high, while data is on bus, then release to latch)
output [15:0] DOUT;	// Data Out to be latched into RAM
output [18:0] ADDRO;	// Address Out, to Memory

reg MEMWRITE;
reg NEWPAC;
reg [15:0] DOUT;
reg [15:0] tempDOUT;	// Holds packet as we buffer it
reg [18:0] ADDRO;
reg [18:0] nextAddress;	// Start writing to memory at this address.  Allows buffering of packets
reg [1:0] counter;
reg [11:0] length;	// This is the overall length to store at the beginning address
reg [2:0] state;	// State of the machine
			
			// Basic flow is:
			// Enable comes on, we start buffering 4 bits into a 16-bit word for memory
			// Watch the "NO" line.  If it goes high while we are working, we can reset to state 0
			// once the enable is low without doing anything further
			// If ENABLE drops, and a "NO" was never encountered, signal that a new good packet
			// has arrived (this is like an IRQ to the controller).  Go back to waiting for Enable
			// You are ensured a 24-cycle gap between packets by the IEEE 802.3 Ethernet Standard
			
			
always @(posedge CLK) begin

	// A "hard" reset, would be a power-on signal on a PCI board
	if(RESET) begin
	 	nextAddress <= 0;
	 	state <= 0;	
	 end
	 
	 else begin
	 
	 case(state)
	 
	 // This is the "wait state"
	 0: begin
	 
	 	 // Finishes up some stuff after a write
	 	length <= 0;
	 	NEWPAC <= 0;
	 	MEMWRITE <= 0;

	 	// Incoming packet, setup to start writing this to memory
	 	if(ENABLE) begin
			ADDRO <= nextAddress;
			tempDOUT[11:8] <= DATAIN; 	// We grab this as the LSBits of the MSBYTE
	 		state <= 1; 			// Start buffering and writing data
	 		counter <= 1;
	 	end

	 	
	 end
	 
	 // This is the buffer and write to memory state; 
	 1: begin
	 
	 	// special case to throw away packet
	 	if(NO) state <= 2;
	 
	 	// Pick the appropriate spot to store data
	 	else begin 
	 	case(counter) 
	 
			0: begin
			 tempDOUT[11:8] <= DATAIN;
			 counter <= 1;
			end
			
			1: begin
			 tempDOUT[15:12] <= DATAIN;
			 counter <= 2;
	 		end
	 		
	 		2: begin
	 		  tempDOUT[3:0] <= DATAIN;
	 		  counter <= 3;
	 		end
	 		
	 		3: begin
	 		  tempDOUT[7:4] <= DATAIN;
	 		  counter <= 0;
	 		end
	 		  
	 	 endcase
	 	 
	 	 // This handles the actual writing on completion and on each halfword
	 	 if(~ENABLE | counter == 0) begin 
	 	 	DOUT <= tempDOUT; 	
	 	 	MEMWRITE <= 1;		// Do the write
	 		length <= length + 1; 	// Increment Length and Address
	 		ADDRO <= ADDRO + 1;
	 		
	 		if(~ENABLE) state <= 3;
	 		 	
	 	 end
	 	 
	 	 else MEMWRITE <= 0; // Drop MEMWRITE after to write to RAM
	 	 		     // This will happen every time counter hits 1
	 	end
	 end
	 	
	 
	 // This is the Toss out the packet state.  Basically we will just write over
	 // the stuff we wrote before seeing "NO".  Not too interesting	 
	 2: if(~ENABLE) state <= 0; // Chill here for a bit
	 
	 // End of packet state, so far it is a success. All that can botch us now
	 // is a failed CRC Check. We also need a pause here to latch data into memory
	 // from the last Nibble
	 3: begin
	 
	    MEMWRITE <= 0; // Drop Memwrite to latch last packet of data
	    // Check for failed CRC
	    if(NO) state <= 2;   
	    else state <= 4;
	    
	    end
	
	 // Write out the length and setup nextAddress    
	 4: begin
	      MEMWRITE <= 1;
	      ADDRO <= nextAddress; // This is where we started from, the end of the prev. packet.
	      nextAddress <= ADDRO; // This is where the next packet should be written. 
	      state <= 0;	    // Go back to waiting for more stuff to do
	      NEWPAC <= 1;
	      DOUT <= length;
	    end
	endcase	 		
  end
end
	 		
endmodule 
