module ETH_PPORT_CTRL(CLK, WE, RE, REQ, DATAIN, DATAOUT, RESET, ADDR, AW, AR, BUSY, MEMIN, MEMOUT, INTERRUPT, NEWPACKET, SEND);

input CLK;
input WE;
input RE;
input [7:0] DATAIN;	// Data from parallel port
input [15:0] MEMIN;	// Data from the memory
input REQ;		// REQ line for handshake
input RESET;
input INTERRUPT;	// Set by receiver to interrupt PPORT for writing a new packet
input SEND;

output [18:0] ADDR;	// 19-bits for SRAM address
output [7:0] DATAOUT;	// Data to parallel port
output [15:0] MEMOUT;	// Data to the memory
output AW, AR; 		// NOTE:  These are inverted
output BUSY;		// For the handshake
output NEWPACKET;	// Signal the MAC that a new packet was received

reg [18:0] READADDR;	// Address where currently reading from
reg [18:0] WRITEADDR;	// Address where currently writing to
reg [7:0] WORD;		// Data buffer
reg AW, AR;		// Mem read & write
reg [15:0] length;    	// Array length (not including start bytes containing this length value)
reg [1:0] state;    	// Overall State Machine Control
reg getLength;		// Specifies whether we are reading the length
reg byteOne;		// Working on first byte of a 16-bit word
reg byteTwo;		// Working on second byte of a 16-bit word
reg stalling;		// Stalling so that the MAC can write to memory
reg AR_HOLD, AW_HOLD;	// For stalls when receiver interrupts, to hold their state
reg NEWPACKET;		// Signal to MAC that we put a new packet in memory
reg [10:0] count;	// Tracks number of bytes read/written from/to parallel port
reg BUSY;
reg [18:0] ADDR;
reg [15:0] MEMOUT;
reg [7:0] DATAOUT;
reg sendok;		// Used to signal arrival and completion of new packet

// Clock based logic
always @(posedge CLK) begin
    // Reset when RESET is high
    if (RESET) begin
       READADDR <= 0;
       WRITEADDR <= 0;
       AW <= 0;
       AR <= 0;
       state <= 0;
       getLength <= 1;
       stalling <= 0;
       byteOne <= 1;
       byteTwo <= 0;
       count <= 0;
       length <= 0;
       NEWPACKET <= 0;
       sendok <= 0;
    end else
    if(~sendok & SEND) sendok <= 1; 
    else
    if(INTERRUPT) begin
      stalling <= 1;
      AR_HOLD <= AR;	// Store values for restoration
      AW_HOLD <= AW;	// Store values for restoration
      AR <= 0;		// Release the buses
      AW <= 0;
    end else
    if(stalling & ~INTERRUPT) begin
      AR <= AR_HOLD;	// Restore previous values
      AW <= AW_HOLD;
      stalling <= 0;	// Not stalling anymore
    end else
    
    // Reset stuff after a full exchange
    if(~WE & ~RE) begin
        AW <= 0;
        AR <= 0;
        state <= 0;
        getLength <= 1;
        count <= 0;
        length <= 0;
        sendok <= 0;
    end else begin
    
    // This is the beef
    case(state)
    0:	begin
    	  if(WE & REQ) begin	// Handle writing
    	      state <= 1;
    	      BUSY <= 1;
    	      count <= count + 1;
    	      
    	      if(byteOne) begin
    	        ADDR <= WRITEADDR;	// Set write address
    	        WORD <= DATAIN;		// Store the first part of the data
    	        if(~getLength)
    	          AW <= 1;		// Write what we got to memory (if not the first time)
    	      end else begin
    	        WRITEADDR <= WRITEADDR + 1;	// Increment write address
    	        MEMOUT <=  {DATAIN, WORD};	// Put it in memory
    	        AW <= 0;			// Release the bus
    	      end
    	      
    	      if(getLength & byteOne)
    	      	length[7:0] <= DATAIN;			// Store the first byte of the length
    	      else if(getLength && byteTwo) begin
    	        length[15:8] <= DATAIN;			// Store the second byte of the length
    	        getLength <= 0;
    	      end
    	      
    	      if((count == length) && (~getLength)) begin	// If at the end, flush it
    	        state <= 2;
    	        NEWPACKET <= 1;
    	        AW <= 1;
    	      end
    	        
    	      byteOne <= byteTwo;	// Switch the byte we're working with
    	      byteTwo <= byteOne;
    	  end
    	  else if(RE & REQ & sendok) begin	// Handle reading
    	      state <= 1;
    	      BUSY <= 1;
    	      count <= count + 1;
    	      
    	      if(byteOne) begin
    	        ADDR <= READADDR;		// Set read address
    	        if(~getLength)			// Show what we got (if not first time)
    	          DATAOUT <= MEMIN[15:8];	// Writeout the second part of the data
    	        AR <= 1;			// Get the word
    	      end else begin
    	        READADDR <= READADDR + 1;	// Increment write address
    	        DATAOUT <=  MEMIN[7:0];		// Write out the first part of the data
    	        AW <= 0;			// Release the bus
    	      end

	      if(getLength & byteTwo) begin
	        length <= MEMIN;
	        getLength <= 0;
	      end
    	      byteOne <= byteTwo;		// Switch the byte we're working with
    	      byteTwo <= byteOne;
    	  end
    	end
    1:  begin
    	  if(!REQ) begin	// Adhere to protocol
    	    state <= 0;
    	    BUSY <= 0;
    	  end
    	end
    2:  begin  // Simple state for flushing to memory and stuff at the end
    	  if(~REQ) begin // Adhere to protocol
    	    AW <= 0;
    	    NEWPACKET <= 0;
    	    BUSY <= 0;
    	  end
    	end
    endcase
end
end
endmodule
