module TRANSMITTER (TXCLK, SEND, SENDOK, RESET, MEM_ADD, LENGTH, SUCCESS, DATAIN, ADDROUT, OE, COLL, DATAOUT, TXEN, CRCENABLE, CRCWRITE) ;

input TXCLK;	// Clock that to obey during XMIT stage

// Incoming & outgoing control signals for external controls
input SEND;		// Asserted by external control when it wants to try to send data
input SENDOK;		// Input from IPG & Random Exponential Backoff timer to OK the xmit
input RESET;		// External reset for state machine
input [18:0]MEM_ADD;	// Start transmitting from here
input [11:0]LENGTH;	// Transmit this number of bytes (note: mem bus is halfword, 2 bytes)
output SUCCESS; 	// Asserted when a transmission is believed to be successful
output CRCENABLE;	// Turns on when CRC should watch data to calc CRC
output CRCWRITE;	// Turns on when CRC's output should contain what we want as the FCS

// Memory Controls (in and out)
input [15:0] DATAIN;	// Halfword from RAM
output [18:0] ADDROUT;	// Load data from this address
output OE;		// Enable RAM for Reading

// Ethernet transmission controls / signals
input COLL;
output [3:0] DATAOUT;
output TXEN;

reg [1:0] state;
reg SUCCESS;
reg ADDROUT;
reg DATAOUT;
reg OE;
reg TXEN;
reg [11:0] counter; // Counts various things at different times: "1010" sent, bytes sent, crc sent
reg [1:0] nibbleCounter; // Keeps track of the number of nibbles sent
reg CRCENABLE;
reg CRCWRITE;

always @(posedge TXCLK) begin

if(COLL) state <= 0;  // At any time, a collision blows the whole transmission, will need to reset

if(RESET) state <= 0;

else begin

	case(state)

	// Waiting for XMIT request	
	0: begin
	
	  if(SEND & SENDOK)begin
	   state <= 2; // Line is clear & ext. control wants to send
	   counter <= 0;
	   TXEN <= 1;
	   ADDROUT <= MEM_ADD;	// Get data from memory
	   OE <= 1;
	  end
	  else begin			// Otherwise, sit right here
		// Don't do anything at all
		SUCCESS <= 0;
		OE <= 0;
		TXEN <= 0;
		CRCENABLE <= 0;
		CRCWRITE <= 0;
	  end
	  
	end
	
	// Transmitting Preamble and SFD "5d"
	1: begin
	
	 if(COLL) state <= 0;  // At any time, a collision blows the whole transmission, will need to reset
	 
	 else if(counter < 14)begin
	 	counter <= counter + 1; // Counter increment
	 	DATAOUT <= 'h5;		// Send the "0101" + SFD 5
	 end	  		  	
	 else begin
	 	counter <= 0;		// Done with counter, reset it for next state
	 	nibbleCounter <= 0;	// Number of nibbles sent (of actual data)
	 	DATAOUT <= 'hd;		// Send the SFD "d"
	 	state <= 2;		// Start xmitting "real" data
	 	ADDROUT <= MEM_ADD;	// Get data from memory
	 	OE <= 1;	
	 end
	end
	// Transmitting actual data (including DEST, SRC, Length/Type, and DATA)
	2: begin
	
	CRCENABLE <= 1;		// Tell the CRC to turn on	
	
	if(COLL) state <= 0;  // At any time, a collision blows the whole transmission, will need to reset
	
	else if(counter < LENGTH) begin
	  case(nibbleCounter)
	  0:begin
	    nibbleCounter <= 1;
	    DATAOUT <= DATAIN[11:8]; 	//MSBYTE, LSBITS
	    end
	    
	  1:begin
	    nibbleCounter <= 2;
	    DATAOUT <= DATAIN[15:12]; 	//MSBYTE, LSBITS
	    counter <= counter + 1;	//A byte has been sent
	    end
	    
	  2:begin
	    nibbleCounter <= 3;
	    DATAOUT <= DATAIN[3:0]; 	//LSBYTE, LSBITS
	    end
	    
	  3:begin
	    nibbleCounter <= 0;
	    DATAOUT <= DATAIN[7:4]; 	//LSBYTE, MSBITS
	    ADDROUT <= ADDROUT + 1;	//Done With HalfWord
	    counter <= counter + 1;	//A 2nd byte has been sent.
	    end   
	
	  endcase	
	end
	
	// Here we handle a length shorter than 46 which is not valid in ethernet
	else if(counter < 46) begin
	  OE <= 0;			// Done with ram, turn it off for other things
	  DATAOUT <= 'h0;			// Write zeros to pad frame
	  
	  // This is a bit tricky, but whenever we come out of the case statement,
	  // nibble will be zero or 2, we only want to increment counter every other time
	  // so we utilize the same 2bit counter
	  if(nibbleCounter[0] == 1) counter <= counter + 1;
	  nibbleCounter <= nibbleCounter + 1;
	end
	
	// We wrote the data and the pad frame if necessary, time to let the CRC do its job
	else begin
	  state <= 3;
	  CRCENABLE <= 0; // Tell the CRC that this is the end, now the output should be the calc'd crc
	  counter <= 0;
	  CRCWRITE <= 1;
	end
	
	end
 	3:begin
 	    if(COLL) state <= 0;  // At any time, a collision blows the whole transmission, will need to reset
 	    else if(counter < 8) counter <= counter + 1;
   	    else state <= 0;	// Done  	  	   	
	end

endcase

end

end // End of always block


endmodule 
